;;; disp-table.el --- functions for dealing with char tables.
;; Keywords: i18n

;; This file is part of XEmacs.

;; XEmacs is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; XEmacs is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with XEmacs; see the file COPYING.  If not, write to the Free
;; Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; #### Need lots of work.  make-display-table depends on a value
;;; that is a define in the C code.  Maybe we should just move the
;;; function into C.

(defun make-display-table ()
  "Return a new, empty display table."
  (make-vector 256 nil))

(provide 'disp-table)

;;; disp-table.el ends here
