/* Synched up with: Not in FSF. */

/* by Hayden Schultz <haydens@ll.mit.edu> for XEmacs */

#ifndef IRIX5_3_H
#define IRIX5_3_H

#include "irix5-1.h"

/*
 * Limit does work in 5.3
 */
#ifdef ULIMIT_BREAK_VALUE
#undef ULIMIT_BREAK_VALUE
#endif

#endif /* IRIX5_1_H */
