/* Synched up with: FSF 19.28. */

/* System description file for hpux version 9.  */

#include "hpux8.h"

#define HPUX9

/* XEmacs change */
#undef C_SWITCH_X_SYSTEM
#define C_SWITCH_X_SYSTEM -I/usr/include/X11R5 -I/usr/include/Motif1.2
#undef LD_SWITCH_X_SYSTEM
#define LD_SWITCH_X_SYSTEM -L/usr/lib/X11R5 -L/usr/lib/Motif1.2

/* If Emacs doesn't seem to work when built to use GNU malloc, you
   probably need to get the latest patches to the HP/UX compiler.
   See `etc/MACHINES' for more information.  */
#if 0
#define SYSTEM_MALLOC 1
#undef GNU_MALLOC
#undef REL_ALLOC
#endif

#ifndef __GNUC__
/* Make room for enough symbols, so dispnew.c does not fail.  */
/* XEmacs: cognot@ensg.u-nancy.fr: C_SWITCH_SYSTEM already defined in hpux8.h,
                           -D_BSD makes hp CC choke on process.c
#define C_SWITCH_SYSTEM -Wp,-H200000 -D_BSD
*/
#undef C_SWITCH_SYSTEM
#define C_SWITCH_SYSTEM -Ae -Wp,-H100000
#endif

/* neal@ctd.comsat.com */
#undef HAVE_TERMIO
#define HAVE_TERMIOS
#define NO_TERMIO

/* XEmacs: apparently rint() is totally broken in HPUX 9. */
#undef HAVE_RINT

